@@ -32,6 +32,8 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "../../src/client/registry.h"
 #include "../../src/client/seat.h"
 #include "../../src/client/shm_pool.h"
+#include "../../src/client/subcompositor.h"
+#include "../../src/client/subsurface.h"
 #include "../../src/client/touch.h"
 #include "../../src/server/buffer_interface.h"
 #include "../../src/server/compositor_interface.h"
@@ -40,11 +42,15 @@ License along with this library.  If not, see <http://www.gnu.org/licenses/>.
 #include "../../src/server/keyboard_interface.h"
 #include "../../src/server/pointer_interface.h"
 #include "../../src/server/seat_interface.h"
+#include "../../src/server/subcompositor_interface.h"
 #include "../../src/server/surface_interface.h"
 // Wayland
 #include <wayland-client-protocol.h>
 
 #include <linux/input.h>
+// System
+#include <fcntl.h>
+#include <unistd.h>
 
 class TestWaylandSeat : public QObject
 {
@@ -63,6 +69,8 @@ private Q_SLOTS:
     void testPointerTransformation();
     void testPointerButton_data();
     void testPointerButton();
+    void testPointerSubSurfaceTree();
+    void testKeyboardSubSurfaceTreeFromPointer();
     void testCursor();
     void testCursorDamage();
     void testKeyboard();
@@ -70,16 +78,19 @@ private Q_SLOTS:
     void testDestroy();
     void testSelection();
     void testTouch();
+    void testDisconnect();
     // TODO: add test for keymap
 
 private:
     KWayland::Server::Display *m_display;
     KWayland::Server::CompositorInterface *m_compositorInterface;
     KWayland::Server::SeatInterface *m_seatInterface;
+    KWayland::Server::SubCompositorInterface *m_subCompositorInterface;
     KWayland::Client::ConnectionThread *m_connection;
     KWayland::Client::Compositor *m_compositor;
     KWayland::Client::Seat *m_seat;
     KWayland::Client::ShmPool *m_shm;
+    KWayland::Client::SubCompositor * m_subCompositor;
     KWayland::Client::EventQueue *m_queue;
     QThread *m_thread;
 };
@@ -91,10 +102,12 @@ TestWaylandSeat::TestWaylandSeat(QObject *parent)
     , m_display(nullptr)
     , m_compositorInterface(nullptr)
     , m_seatInterface(nullptr)
+    , m_subCompositorInterface(nullptr)
     , m_connection(nullptr)
     , m_compositor(nullptr)
     , m_seat(nullptr)
     , m_shm(nullptr)
+    , m_subCompositor(nullptr)
     , m_queue(nullptr)
     , m_thread(nullptr)
 {
@@ -115,6 +128,11 @@ void TestWaylandSeat::init()
     m_compositorInterface->create();
     QVERIFY(m_compositorInterface->isValid());
 
+    m_subCompositorInterface = m_display->createSubCompositor(m_display);
+    QVERIFY(m_subCompositorInterface);
+    m_subCompositorInterface->create();
+    QVERIFY(m_subCompositorInterface->isValid());
+
     // setup connection
     m_connection = new KWayland::Client::ConnectionThread;
     QSignalSpy connectedSpy(m_connection, SIGNAL(connected()));
@@ -158,10 +176,19 @@ void TestWaylandSeat::init()
     m_shm = new KWayland::Client::ShmPool(this);
     m_shm->setup(registry.bindShm(shmSpy.first().first().value<quint32>(), shmSpy.first().last().value<quint32>()));
     QVERIFY(m_shm->isValid());
+
+    m_subCompositor = registry.createSubCompositor(registry.interface(KWayland::Client::Registry::Interface::SubCompositor).name,
+                                                   registry.interface(KWayland::Client::Registry::Interface::SubCompositor).version,
+                                                   this);
+    QVERIFY(m_subCompositor->isValid());
 }
 
 void TestWaylandSeat::cleanup()
 {
+    if (m_subCompositor) {
+        delete m_subCompositor;
+        m_subCompositor = nullptr;
+    }
     if (m_shm) {
         delete m_shm;
         m_shm = nullptr;
@@ -195,6 +222,9 @@ void TestWaylandSeat::cleanup()
     delete m_seatInterface;
     m_seatInterface = nullptr;
 
+    delete m_subCompositorInterface;
+    m_subCompositorInterface = nullptr;
+
     delete m_display;
     m_display = nullptr;
 }
@@ -436,11 +466,56 @@ void TestWaylandSeat::testPointer()
     QCOMPARE(p->enteredSurface(), s);
     QCOMPARE(cp.enteredSurface(), s);
 
+    // destroy the focused pointer
+    QSignalSpy unboundSpy(serverPointer, &Resource::unbound);
+    QVERIFY(unboundSpy.isValid());
+    QSignalSpy destroyedSpy(serverPointer, &Resource::destroyed);
+    QVERIFY(destroyedSpy.isValid());
+    delete p;
+    QVERIFY(unboundSpy.wait());
+    QCOMPARE(unboundSpy.count(), 1);
+    QCOMPARE(destroyedSpy.count(), 0);
+    // now test that calling into the methods in Seat does not crash
+    QCOMPARE(m_seatInterface->focusedPointer(), serverPointer);
+    QCOMPARE(m_seatInterface->focusedPointerSurface(), serverSurface);
+    m_seatInterface->setTimestamp(8);
+    m_seatInterface->setPointerPos(QPoint(10, 15));
+    m_seatInterface->setTimestamp(9);
+    m_seatInterface->pointerButtonPressed(1);
+    m_seatInterface->setTimestamp(10);
+    m_seatInterface->pointerButtonReleased(1);
+    m_seatInterface->setTimestamp(11);
+    m_seatInterface->pointerAxis(Qt::Horizontal, 10);
+    m_seatInterface->setTimestamp(12);
+    m_seatInterface->pointerAxis(Qt::Vertical, 20);
+    m_seatInterface->setFocusedPointerSurface(nullptr);
+    QCOMPARE(focusedPointerChangedSpy.count(), 7);
+    m_seatInterface->setFocusedPointerSurface(serverSurface);
+    QCOMPARE(focusedPointerChangedSpy.count(), 8);
+    QCOMPARE(m_seatInterface->focusedPointerSurface(), serverSurface);
+    QVERIFY(!m_seatInterface->focusedPointer());
+
+    // and now destroy
+    QVERIFY(destroyedSpy.wait());
+    QCOMPARE(unboundSpy.count(), 1);
+    QCOMPARE(destroyedSpy.count(), 1);
+    QCOMPARE(m_seatInterface->focusedPointerSurface(), serverSurface);
+    QVERIFY(!m_seatInterface->focusedPointer());
+
+    // create a pointer again
+    p = m_seat->createPointer(m_seat);
+    QVERIFY(focusedPointerChangedSpy.wait());
+    QCOMPARE(focusedPointerChangedSpy.count(), 9);
+    QCOMPARE(m_seatInterface->focusedPointerSurface(), serverSurface);
+    serverPointer = m_seatInterface->focusedPointer();
+    QVERIFY(serverPointer);
+
     delete s;
     wl_display_flush(m_connection->display());
     QVERIFY(focusedPointerChangedSpy.wait());
-    QCOMPARE(focusedPointerChangedSpy.count(), 7);
+    QCOMPARE(focusedPointerChangedSpy.count(), 10);
     QVERIFY(!m_seatInterface->focusedPointerSurface());
+    QVERIFY(!m_seatInterface->focusedPointer());
 }
 
 void TestWaylandSeat::testPointerTransformation_data()
@@ -643,6 +718,206 @@ void TestWaylandSeat::testPointerButton()
     QCOMPARE(buttonChangedSpy.last().at(3).value<KWayland::Client::Pointer::ButtonState>(), Pointer::ButtonState::Released);
 }
 
+void TestWaylandSeat::testPointerSubSurfaceTree()
+{
+    // this test verifies that pointer motion on a surface with sub-surfaces sends motion enter/leave to the sub-surface
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+
+    // first create the pointer
+    QSignalSpy hasPointerChangedSpy(m_seat, &Seat::hasPointerChanged);
+    QVERIFY(hasPointerChangedSpy.isValid());
+    m_seatInterface->setHasPointer(true);
+    QVERIFY(hasPointerChangedSpy.wait());
+    QScopedPointer<Pointer> pointer(m_seat->createPointer());
+
+    // create a sub surface tree
+    // parent surface (100, 100) with one sub surface taking the half of it's size (50, 100)
+    // which has two further children (50, 50) which are overlapping
+    QSignalSpy surfaceCreatedSpy(m_compositorInterface, &CompositorInterface::surfaceCreated);
+    QVERIFY(surfaceCreatedSpy.isValid());
+    QScopedPointer<Surface> parentSurface(m_compositor->createSurface());
+    QScopedPointer<Surface> childSurface(m_compositor->createSurface());
+    QScopedPointer<Surface> grandChild1Surface(m_compositor->createSurface());
+    QScopedPointer<Surface> grandChild2Surface(m_compositor->createSurface());
+    QScopedPointer<SubSurface> childSubSurface(m_subCompositor->createSubSurface(childSurface.data(), parentSurface.data()));
+    QScopedPointer<SubSurface> grandChild1SubSurface(m_subCompositor->createSubSurface(grandChild1Surface.data(), childSurface.data()));
+    QScopedPointer<SubSurface> grandChild2SubSurface(m_subCompositor->createSubSurface(grandChild2Surface.data(), childSurface.data()));
+    grandChild2SubSurface->setPosition(QPoint(0, 25));
+
+    // let's map the surfaces
+    auto render = [this] (Surface *s, const QSize &size) {
+        QImage image(size, QImage::Format_ARGB32);
+        image.fill(Qt::black);
+        s->attachBuffer(m_shm->createBuffer(image));
+        s->damage(QRect(QPoint(0, 0), size));
+        s->commit(Surface::CommitFlag::None);
+    };
+    render(grandChild2Surface.data(), QSize(50, 50));
+    render(grandChild1Surface.data(), QSize(50, 50));
+    render(childSurface.data(), QSize(50, 100));
+    render(parentSurface.data(), QSize(100, 100));
+
+    QVERIFY(surfaceCreatedSpy.wait());
+    auto serverSurface = surfaceCreatedSpy.first().first().value<SurfaceInterface*>();
+    QVERIFY(serverSurface->isMapped());
+
+    // send in pointer events
+    QSignalSpy enteredSpy(pointer.data(), &Pointer::entered);
+    QVERIFY(enteredSpy.isValid());
+    QSignalSpy leftSpy(pointer.data(), &Pointer::left);
+    QVERIFY(leftSpy.isValid());
+    QSignalSpy motionSpy(pointer.data(), &Pointer::motion);
+    QVERIFY(motionSpy.isValid());
+    // first to the grandChild2 in the overlapped area
+    quint32 timestamp = 1;
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->setPointerPos(QPointF(25, 50));
+    m_seatInterface->setFocusedPointerSurface(serverSurface);
+    QVERIFY(enteredSpy.wait());
+    QCOMPARE(enteredSpy.count(), 1);
+    QCOMPARE(leftSpy.count(), 0);
+    QCOMPARE(motionSpy.count(), 0);
+    QCOMPARE(enteredSpy.last().last().toPointF(), QPointF(25, 25));
+    QCOMPARE(pointer->enteredSurface(), grandChild2Surface.data());
+    // a motion on grandchild2
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->setPointerPos(QPointF(25, 60));
+    QVERIFY(motionSpy.wait());
+    QCOMPARE(enteredSpy.count(), 1);
+    QCOMPARE(leftSpy.count(), 0);
+    QCOMPARE(motionSpy.count(), 1);
+    QCOMPARE(motionSpy.last().first().toPointF(), QPointF(25, 35));
+    // motion which changes to childSurface
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->setPointerPos(QPointF(25, 80));
+    QVERIFY(enteredSpy.wait());
+    QCOMPARE(enteredSpy.count(), 2);
+    QCOMPARE(leftSpy.count(), 1);
+    QCOMPARE(motionSpy.count(), 1);
+    QCOMPARE(enteredSpy.last().last().toPointF(), QPointF(25, 80));
+    QCOMPARE(pointer->enteredSurface(), childSurface.data());
+    // a leave for the whole surface
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->setFocusedPointerSurface(nullptr);
+    QVERIFY(leftSpy.wait());
+    QCOMPARE(enteredSpy.count(), 2);
+    QCOMPARE(leftSpy.count(), 2);
+    QCOMPARE(motionSpy.count(), 1);
+    // a new enter on the main surface
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->setPointerPos(QPointF(75, 50));
+    m_seatInterface->setFocusedPointerSurface(serverSurface);
+    QVERIFY(enteredSpy.wait());
+    QCOMPARE(enteredSpy.count(), 3);
+    QCOMPARE(leftSpy.count(), 2);
+    QCOMPARE(motionSpy.count(), 1);
+    QCOMPARE(enteredSpy.last().last().toPointF(), QPointF(75, 50));
+    QCOMPARE(pointer->enteredSurface(), parentSurface.data());
+}
+
+void TestWaylandSeat::testKeyboardSubSurfaceTreeFromPointer()
+{
+    // this test verifies that when clicking on a sub-surface the keyboard focus passes to it
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+
+    // first create the pointer
+    QSignalSpy hasPointerChangedSpy(m_seat, &Seat::hasPointerChanged);
+    QVERIFY(hasPointerChangedSpy.isValid());
+    m_seatInterface->setHasPointer(true);
+    QVERIFY(hasPointerChangedSpy.wait());
+    QScopedPointer<Pointer> pointer(m_seat->createPointer());
+
+    // and create keyboard
+    QSignalSpy hasKeyboardChangedSpy(m_seat, &Seat::hasKeyboardChanged);
+    QVERIFY(hasKeyboardChangedSpy.isValid());
+    m_seatInterface->setHasKeyboard(true);
+    QVERIFY(hasKeyboardChangedSpy.wait());
+    QScopedPointer<Keyboard> keyboard(m_seat->createKeyboard());
+
+    // create a sub surface tree
+    // parent surface (100, 100) with one sub surface taking the half of it's size (50, 100)
+    // which has two further children (50, 50) which are overlapping
+    QSignalSpy surfaceCreatedSpy(m_compositorInterface, &CompositorInterface::surfaceCreated);
+    QVERIFY(surfaceCreatedSpy.isValid());
+    QScopedPointer<Surface> parentSurface(m_compositor->createSurface());
+    QScopedPointer<Surface> childSurface(m_compositor->createSurface());
+    QScopedPointer<Surface> grandChild1Surface(m_compositor->createSurface());
+    QScopedPointer<Surface> grandChild2Surface(m_compositor->createSurface());
+    QScopedPointer<SubSurface> childSubSurface(m_subCompositor->createSubSurface(childSurface.data(), parentSurface.data()));
+    QScopedPointer<SubSurface> grandChild1SubSurface(m_subCompositor->createSubSurface(grandChild1Surface.data(), childSurface.data()));
+    QScopedPointer<SubSurface> grandChild2SubSurface(m_subCompositor->createSubSurface(grandChild2Surface.data(), childSurface.data()));
+    grandChild2SubSurface->setPosition(QPoint(0, 25));
+
+    // let's map the surfaces
+    auto render = [this] (Surface *s, const QSize &size) {
+        QImage image(size, QImage::Format_ARGB32);
+        image.fill(Qt::black);
+        s->attachBuffer(m_shm->createBuffer(image));
+        s->damage(QRect(QPoint(0, 0), size));
+        s->commit(Surface::CommitFlag::None);
+    };
+    render(grandChild2Surface.data(), QSize(50, 50));
+    render(grandChild1Surface.data(), QSize(50, 50));
+    render(childSurface.data(), QSize(50, 100));
+    render(parentSurface.data(), QSize(100, 100));
+
+    QVERIFY(surfaceCreatedSpy.wait());
+    auto serverSurface = surfaceCreatedSpy.first().first().value<SurfaceInterface*>();
+    QVERIFY(serverSurface->isMapped());
+
+    // pass keyboard focus to the main surface
+    QSignalSpy enterSpy(keyboard.data(), &Keyboard::entered);
+    QVERIFY(enterSpy.isValid());
+    QSignalSpy leftSpy(keyboard.data(), &Keyboard::left);
+    QVERIFY(leftSpy.isValid());
+    m_seatInterface->setFocusedKeyboardSurface(serverSurface);
+    QVERIFY(enterSpy.wait());
+    QCOMPARE(enterSpy.count(), 1);
+    QCOMPARE(leftSpy.count(), 0);
+    QCOMPARE(keyboard->enteredSurface(), parentSurface.data());
+
+    // now pass also pointer focus to the surface
+    QSignalSpy pointerEnterSpy(pointer.data(), &Pointer::entered);
+    QVERIFY(pointerEnterSpy.isValid());
+    quint32 timestamp = 1;
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->setPointerPos(QPointF(25, 50));
+    m_seatInterface->setFocusedPointerSurface(serverSurface);
+    QVERIFY(pointerEnterSpy.wait());
+    QCOMPARE(pointerEnterSpy.count(), 1);
+    // should not have affected the keyboard
+    QCOMPARE(enterSpy.count(), 1);
+    QCOMPARE(leftSpy.count(), 0);
+
+    // let's click
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->pointerButtonPressed(Qt::LeftButton);
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->pointerButtonReleased(Qt::LeftButton);
+    QVERIFY(enterSpy.wait());
+    QCOMPARE(enterSpy.count(), 2);
+    QCOMPARE(leftSpy.count(), 1);
+    QCOMPARE(keyboard->enteredSurface(), grandChild2Surface.data());
+
+    // click on same surface should not trigger another enter
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->pointerButtonPressed(Qt::LeftButton);
+    m_seatInterface->setTimestamp(timestamp++);
+    m_seatInterface->pointerButtonReleased(Qt::LeftButton);
+    QVERIFY(!enterSpy.wait(200));
+    QCOMPARE(enterSpy.count(), 2);
+    QCOMPARE(leftSpy.count(), 1);
+    QCOMPARE(keyboard->enteredSurface(), grandChild2Surface.data());
+
+    // unfocus keyboard
+    m_seatInterface->setFocusedKeyboardSurface(nullptr);
+    QVERIFY(leftSpy.wait());
+    QCOMPARE(enterSpy.count(), 2);
+    QCOMPARE(leftSpy.count(), 2);
+}
+
 void TestWaylandSeat::testCursor()
 {
     using namespace KWayland::Client;
@@ -688,6 +963,7 @@ void TestWaylandSeat::testCursor()
     QVERIFY(!cursor->surface());
     QCOMPARE(cursor->hotspot(), QPoint());
     QCOMPARE(cursor->enteredSerial(), serial);
+    QCOMPARE(cursor->pointer(), m_seatInterface->focusedPointer());
 
     QSignalSpy hotspotChangedSpy(cursor, SIGNAL(hotspotChanged()));
     QVERIFY(hotspotChangedSpy.isValid());
@@ -839,7 +1115,8 @@ void TestWaylandSeat::testKeyboard()
     QCOMPARE(keyboard->keyRepeatRate(), 0);
     wl_display_flush(m_connection->display());
     QTest::qWait(100);
-    QVERIFY(m_seatInterface->focusedKeyboard());
+    auto serverKeyboard = m_seatInterface->focusedKeyboard();
+    QVERIFY(serverKeyboard);
 
     // we should get the repeat info announced
     QCOMPARE(repeatInfoSpy.count(), 1);
@@ -880,7 +1157,8 @@ void TestWaylandSeat::testKeyboard()
     QCOMPARE(modifierSpy.first().at(2).value<quint32>(), quint32(0));
     QCOMPARE(modifierSpy.first().at(3).value<quint32>(), quint32(0));
     QCOMPARE(enteredSpy.count(), 1);
-    QCOMPARE(enteredSpy.first().first().value<quint32>(), m_display->serial());
+    // TODO: get through API
+    QCOMPARE(enteredSpy.first().first().value<quint32>(), m_display->serial() - 1);
 
     QSignalSpy keyChangedSpy(keyboard, SIGNAL(keyChanged(quint32,KWayland::Client::Keyboard::KeyState,quint32)));
     QVERIFY(keyChangedSpy.isValid());
@@ -933,7 +1211,8 @@ void TestWaylandSeat::testKeyboard()
     QVERIFY(!m_seatInterface->focusedKeyboard());
     QVERIFY(leftSpy.wait());
     QCOMPARE(leftSpy.count(), 1);
-    QCOMPARE(leftSpy.first().first().value<quint32>(), m_display->serial());
+    // TODO: get through API
+    QCOMPARE(leftSpy.first().first().value<quint32>(), m_display->serial() -1 );
 
     QVERIFY(!keyboard->enteredSurface());
     QVERIFY(!ckeyboard.enteredSurface());
@@ -948,11 +1227,49 @@ void TestWaylandSeat::testKeyboard()
     QCOMPARE(keyboard->enteredSurface(), s);
     QCOMPARE(ckeyboard.enteredSurface(), s);
 
+    QSignalSpy serverSurfaceDestroyedSpy(serverSurface, &QObject::destroyed);
+    QVERIFY(serverSurfaceDestroyedSpy.isValid());
     delete s;
-    wl_display_flush(m_connection->display());
-    QTest::qWait(100);
+    QVERIFY(serverSurfaceDestroyedSpy.wait());
     QVERIFY(!m_seatInterface->focusedKeyboardSurface());
     QVERIFY(!m_seatInterface->focusedKeyboard());
+    QVERIFY(!serverKeyboard->focusedSurface());
+
+    // let's create a Surface again
+    QScopedPointer<Surface> s2(m_compositor->createSurface());
+    QVERIFY(surfaceCreatedSpy.wait());
+    QCOMPARE(surfaceCreatedSpy.count(), 2);
+    serverSurface = surfaceCreatedSpy.last().first().value<SurfaceInterface*>();
+    QVERIFY(serverSurface);
+    m_seatInterface->setFocusedKeyboardSurface(serverSurface);
+    QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
+    QCOMPARE(m_seatInterface->focusedKeyboard(), serverKeyboard);
+
+    // delete the Keyboard
+    QSignalSpy unboundSpy(serverKeyboard, &Resource::unbound);
+    QVERIFY(unboundSpy.isValid());
+    QSignalSpy destroyedSpy(serverKeyboard, &Resource::destroyed);
+    QVERIFY(destroyedSpy.isValid());
+    delete keyboard;
+    QVERIFY(unboundSpy.wait());
+    QCOMPARE(unboundSpy.count(), 1);
+    QCOMPARE(destroyedSpy.count(), 0);
+    // verify that calling into the Keyboard related functionality doesn't crash
+    m_seatInterface->setTimestamp(9);
+    m_seatInterface->keyPressed(KEY_F2);
+    m_seatInterface->setTimestamp(10);
+    m_seatInterface->keyReleased(KEY_F2);
+    m_seatInterface->setKeyRepeatInfo(30, 560);
+    m_seatInterface->setKeyRepeatInfo(25, 660);
+    m_seatInterface->updateKeyboardModifiers(5, 6, 7, 8);
+    m_seatInterface->setKeymap(open("/dev/null", O_RDONLY), 0);
+    m_seatInterface->setFocusedKeyboardSurface(nullptr);
+    m_seatInterface->setFocusedKeyboardSurface(serverSurface);
+    QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
+    QVERIFY(!m_seatInterface->focusedKeyboard());
+
+    QVERIFY(destroyedSpy.wait());
+    QCOMPARE(destroyedSpy.count(), 1);
 
     // create a second Keyboard to verify that repeat info is announced properly
     Keyboard *keyboard2 = m_seat->createKeyboard(m_seat);
@@ -967,6 +1284,17 @@ void TestWaylandSeat::testKeyboard()
     QCOMPARE(keyboard2->isKeyRepeatEnabled(), true);
     QCOMPARE(keyboard2->keyRepeatRate(), 25);
     QCOMPARE(keyboard2->keyRepeatDelay(), 660);
+    QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
+    serverKeyboard = m_seatInterface->focusedKeyboard();
+    QVERIFY(serverKeyboard);
+    QSignalSpy keyboard2DestroyedSpy(serverKeyboard, &QObject::destroyed);
+    QVERIFY(keyboard2DestroyedSpy.isValid());
+    delete keyboard2;
+    QVERIFY(keyboard2DestroyedSpy.wait());
+    // this should have unset it on the server
+    QVERIFY(!m_seatInterface->focusedKeyboard());
+    // but not the surface
+    QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
 }
 
 void TestWaylandSeat::testCast()
@@ -1019,6 +1347,7 @@ void TestWaylandSeat::testDestroy()
     m_compositor = nullptr;
     connect(m_connection, &ConnectionThread::connectionDied, m_seat, &Seat::destroy);
     connect(m_connection, &ConnectionThread::connectionDied, m_shm, &ShmPool::destroy);
+    connect(m_connection, &ConnectionThread::connectionDied, m_subCompositor, &SubCompositor::destroy);
     connect(m_connection, &ConnectionThread::connectionDied, m_queue, &EventQueue::destroy);
     QVERIFY(m_seat->isValid());
 
@@ -1028,6 +1357,7 @@ void TestWaylandSeat::testDestroy()
     m_display = nullptr;
     m_compositorInterface = nullptr;
     m_seatInterface = nullptr;
+    m_subCompositorInterface = nullptr;
     QVERIFY(connectionDiedSpy.wait());
 
     // now the seat should be destroyed;
@@ -1075,6 +1405,7 @@ void TestWaylandSeat::testSelection()
     QVERIFY(surface->isValid());
     QVERIFY(surfaceCreatedSpy.wait());
     auto serverSurface = surfaceCreatedSpy.first().first().value<SurfaceInterface*>();
+    QVERIFY(!m_seatInterface->selection());
     m_seatInterface->setFocusedKeyboardSurface(serverSurface);
     QCOMPARE(m_seatInterface->focusedKeyboardSurface(), serverSurface);
     QVERIFY(!m_seatInterface->focusedKeyboard());
@@ -1083,6 +1414,7 @@ void TestWaylandSeat::testSelection()
     QCoreApplication::processEvents();
     QVERIFY(selectionSpy.isEmpty());
     QVERIFY(selectionClearedSpy.isEmpty());
+    QVERIFY(!m_seatInterface->selection());
 
     // now let's try to set a selection - we have keyboard focus, so it should be sent to us
     QScopedPointer<DataSource> ds(ddm->createDataSource());
@@ -1091,6 +1423,8 @@ void TestWaylandSeat::testSelection()
     dd1->setSelection(0, ds.data());
     QVERIFY(selectionSpy.wait());
     QCOMPARE(selectionSpy.count(), 1);
+    auto ddi = m_seatInterface->selection();
+    QVERIFY(ddi);
     QVERIFY(selectionClearedSpy.isEmpty());
     auto df = selectionSpy.first().first().value<DataOffer*>();
     QCOMPARE(df->offeredMimeTypes().count(), 1);
@@ -1116,6 +1450,25 @@ void TestWaylandSeat::testSelection()
     QCoreApplication::processEvents();
     QCoreApplication::processEvents();
     QCOMPARE(selectionSpy.count(), 1);
+
+    // let's unset the selection on the seat
+    m_seatInterface->setSelection(nullptr);
+    // and pass focus back on our surface
+    m_seatInterface->setFocusedKeyboardSurface(serverSurface);
+    // we don't have a selection, so it should not send a selection
+    QVERIFY(!selectionSpy.wait(100));
+    // now let's set it manually
+    m_seatInterface->setSelection(ddi);
+    QCOMPARE(m_seatInterface->selection(), ddi);
+    QVERIFY(selectionSpy.wait());
+    QCOMPARE(selectionSpy.count(), 2);
+    // setting the same again should not change
+    m_seatInterface->setSelection(ddi);
+    QVERIFY(!selectionSpy.wait(100));
+    // now clear it manully
+    m_seatInterface->setSelection(nullptr);
+    QVERIFY(selectionClearedSpy.wait());
+    QCOMPARE(selectionSpy.count(), 2);
 }
 
 void TestWaylandSeat::testTouch()
@@ -1146,7 +1499,8 @@ void TestWaylandSeat::testTouch()
     Touch *touch = m_seat->createTouch(m_seat);
     QVERIFY(touch->isValid());
     QVERIFY(touchCreatedSpy.wait());
-    QVERIFY(m_seatInterface->focusedTouch());
+    auto serverTouch = m_seatInterface->focusedTouch();
+    QVERIFY(serverTouch);
     QCOMPARE(touchCreatedSpy.first().first().value<KWayland::Server::TouchInterface*>(), m_seatInterface->focusedTouch());
 
     QSignalSpy sequenceStartedSpy(touch, SIGNAL(sequenceStarted(KWayland::Client::TouchPoint*)));
@@ -1305,6 +1659,103 @@ void TestWaylandSeat::testTouch()
     QCOMPARE(pointMovedSpy.count(), 1);
     QCOMPARE(pointRemovedSpy.count(), 3);
     QCOMPARE(touch->sequence().first()->position(), QPointF(0, 0));
+
+    // destroy touch on client side
+    QSignalSpy unboundSpy(serverTouch, &TouchInterface::unbound);
+    QVERIFY(unboundSpy.isValid());
+    QSignalSpy destroyedSpy(serverTouch, &TouchInterface::destroyed);
+    QVERIFY(destroyedSpy.isValid());
+    delete touch;
+    QVERIFY(unboundSpy.wait());
+    QCOMPARE(unboundSpy.count(), 1);
+    QCOMPARE(destroyedSpy.count(), 0);
+    QVERIFY(!serverTouch->resource());
+    // try to call into all the the methods of the touch interface, should not crash
+    QCOMPARE(m_seatInterface->focusedTouch(), serverTouch);
+    m_seatInterface->setTimestamp(8);
+    QCOMPARE(m_seatInterface->touchDown(QPointF(15, 26)), 0);
+    m_seatInterface->touchFrame();
+    m_seatInterface->touchMove(0, QPointF(0, 0));
+    QCOMPARE(m_seatInterface->touchDown(QPointF(15, 26)), 1);
+    m_seatInterface->cancelTouchSequence();
+    QVERIFY(destroyedSpy.wait());
+    QCOMPARE(destroyedSpy.count(), 1);
+    // should have unset the focused touch
+    QVERIFY(!m_seatInterface->focusedTouch());
+    // but not the focused touch surface
+    QCOMPARE(m_seatInterface->focusedTouchSurface(), serverSurface);
+}
+
+void TestWaylandSeat::testDisconnect()
+{
+    // this test verifies that disconnecting the client cleans up correctly
+    using namespace KWayland::Client;
+    using namespace KWayland::Server;
+    QSignalSpy keyboardCreatedSpy(m_seatInterface, &SeatInterface::keyboardCreated);
+    QVERIFY(keyboardCreatedSpy.isValid());
+    QSignalSpy pointerCreatedSpy(m_seatInterface, &SeatInterface::pointerCreated);
+    QVERIFY(pointerCreatedSpy.isValid());
+    QSignalSpy touchCreatedSpy(m_seatInterface, &SeatInterface::touchCreated);
+    QVERIFY(touchCreatedSpy.isValid());
+
+    // create the things we need
+    m_seatInterface->setHasKeyboard(true);
+    m_seatInterface->setHasPointer(true);
+    m_seatInterface->setHasTouch(true);
+    QSignalSpy touchSpy(m_seat, &Seat::hasTouchChanged);
+    QVERIFY(touchSpy.isValid());
+    QVERIFY(touchSpy.wait());
+
+    QScopedPointer<Keyboard> keyboard(m_seat->createKeyboard());
+    QVERIFY(!keyboard.isNull());
+    QVERIFY(keyboardCreatedSpy.wait());
+    auto serverKeyboard = keyboardCreatedSpy.first().first().value<KeyboardInterface*>();
+    QVERIFY(serverKeyboard);
+
+    QScopedPointer<Pointer> pointer(m_seat->createPointer());
+    QVERIFY(!pointer.isNull());
+    QVERIFY(pointerCreatedSpy.wait());
+    auto serverPointer = pointerCreatedSpy.first().first().value<PointerInterface*>();
+    QVERIFY(serverPointer);
+
+    QScopedPointer<Touch> touch(m_seat->createTouch());
+    QVERIFY(!touch.isNull());
+    QVERIFY(touchCreatedSpy.wait());
+    auto serverTouch = touchCreatedSpy.first().first().value<TouchInterface*>();
+    QVERIFY(serverTouch);
+
+    // setup destroys
+    QSignalSpy keyboardDestroyedSpy(serverKeyboard, &QObject::destroyed);
+    QVERIFY(keyboardDestroyedSpy.isValid());
+    QSignalSpy pointerDestroyedSpy(serverPointer, &QObject::destroyed);
+    QVERIFY(pointerDestroyedSpy.isValid());
+    QSignalSpy touchDestroyedSpy(serverTouch, &QObject::destroyed);
+    QVERIFY(touchDestroyedSpy.isValid());
+    QSignalSpy clientDisconnectedSpy(serverKeyboard->client(), &ClientConnection::disconnected);
+    QVERIFY(clientDisconnectedSpy.isValid());
+
+    if (m_connection) {
+        m_connection->deleteLater();
+        m_connection = nullptr;
+    }
+    QVERIFY(clientDisconnectedSpy.wait());
+    QCOMPARE(clientDisconnectedSpy.count(), 1);
+    QCOMPARE(keyboardDestroyedSpy.count(), 0);
+    QCOMPARE(pointerDestroyedSpy.count(), 0);
+    QCOMPARE(touchDestroyedSpy.count(), 0);
+    QVERIFY(keyboardDestroyedSpy.wait());
+    QCOMPARE(keyboardDestroyedSpy.count(), 1);
+    QCOMPARE(pointerDestroyedSpy.count(), 1);
+    QCOMPARE(touchDestroyedSpy.count(), 1);
+
+    keyboard->destroy();
+    pointer->destroy();
+    touch->destroy();
+    m_compositor->destroy();
+    m_seat->destroy();
+    m_shm->destroy();
+    m_subCompositor->destroy();
+    m_queue->destroy();
 }
 
 QTEST_GUILESS_MAIN(TestWaylandSeat)
